@@ -50,10 +50,7 @@
 /* #define DEBUG_STACK_SIZE */
 
 /* I/O buffer size */
-#define BUFFSIZE 16384
-
-/* IP address and TCP port textual representation length */
-#define IPLEN 128
+#define BUFFSIZE 8192
 
 /* how many bytes of random input to read from files for PRNG */
 /* OpenSSL likes at least 128 bits, so 64 bytes seems plenty. */
@@ -64,25 +61,68 @@
 
 /**************************************** platform */
 
-#ifdef USE_WIN32
-#define USE_IPv6
-/* #define USE_FIPS */
+#ifdef _WIN32
+#define USE_WIN32
 #endif
 
 #ifdef _WIN32_WCE
 #define USE_WIN32
 typedef int socklen_t;
-#define EINTR WSAEINTR
-#define EMFILE WSAEMFILE
 #endif
 
 #ifdef USE_WIN32
-#define HAVE_OPENSSL
+#define USE_IPv6
+/* #define USE_FIPS */
+#define _CRT_SECURE_NO_DEPRECATE
+#define _CRT_NONSTDC_NO_DEPRECATE
 #define HAVE_OSSL_ENGINE_H
+#define HAVE_OSSL_OCSP_H
 /* prevent including wincrypt.h, as it defines it's own OCSP_RESPONSE */
 #define __WINCRYPT_H__
 #endif
 
+#ifdef USE_WIN32
+#define S_EADDRINUSE  WSAEADDRINUSE
+/* winsock does not define WSAEAGAIN */
+/* in most (but not all!) BSD implementations EAGAIN==EWOULDBLOCK */
+#define S_EAGAIN      WSAEWOULDBLOCK
+#define S_ECONNRESET  WSAECONNRESET
+#define S_EINPROGRESS WSAEINPROGRESS
+#define S_EINTR       WSAEINTR
+#define S_EINVAL      WSAEINVAL
+#define S_EISCONN     WSAEISCONN
+#define S_EMFILE      WSAEMFILE
+/* winsock does not define WSAENFILE */
+#define S_ENOBUFS     WSAENOBUFS
+/* winsock does not define WSAENOMEM */
+#define S_ENOPROTOOPT WSAENOPROTOOPT
+#define S_ENOTSOCK    WSAENOTSOCK
+#define S_EOPNOTSUPP  WSAEOPNOTSUPP
+#define S_EWOULDBLOCK WSAEWOULDBLOCK
+#else /* USE_WIN32 */
+#define S_EADDRINUSE  EADDRINUSE
+#define S_EAGAIN      EAGAIN
+#define S_ECONNRESET  ECONNRESET
+#define S_EINPROGRESS EINPROGRESS
+#define S_EINTR       EINTR
+#define S_EINVAL      EINVAL
+#define S_EISCONN     EISCONN
+#define S_EMFILE      EMFILE
+#ifdef ENFILE
+#define S_ENFILE      ENFILE
+#endif
+#ifdef ENOBUFS
+#define S_ENOBUFS     ENOBUFS
+#endif
+#ifdef ENOMEM
+#define S_ENOMEM      ENOMEM
+#endif
+#define S_ENOPROTOOPT ENOPROTOOPT
+#define S_ENOTSOCK    ENOTSOCK
+#define S_EOPNOTSUPP  EOPNOTSUPP
+#define S_EWOULDBLOCK EWOULDBLOCK
+#endif /* USE_WIN32 */
+
 /**************************************** generic headers */
 
 #ifdef __vms
@@ -105,6 +145,7 @@ typedef int socklen_t;
 #define THREADS
 #endif
 #ifndef _REENTRANT
+/* _REENTRANT is required for thread-safe errno on Solaris */
 #define _REENTRANT
 #endif
 #ifndef _THREAD_SAFE
@@ -164,27 +205,6 @@ typedef unsigned long u32;
 #include <ws2tcpip.h>
 #include <windows.h>
 
-/* needed for winsock2 (not for winsock) */
-#define ENOTSOCK WSAENOTSOCK
-#define ENOPROTOOPT WSAENOPROTOOPT
-#define EINPROGRESS WSAEINPROGRESS
-#define EWOULDBLOCK WSAEWOULDBLOCK
-#define EADDRINUSE WSAEADDRINUSE
-
-#define ECONNRESET WSAECONNRESET
-#define ENOTSOCK WSAENOTSOCK
-#define ENOPROTOOPT WSAENOPROTOOPT
-#define EINPROGRESS WSAEINPROGRESS
-#define EWOULDBLOCK WSAEWOULDBLOCK
-#define EISCONN WSAEISCONN
-#define EADDRINUSE WSAEADDRINUSE
-#define EOPNOTSUPP WSAEOPNOTSUPP
-
-#ifdef EINVAL
-#undef EINVAL
-#endif
-#define EINVAL WSAEINVAL
-
 #include <process.h>     /* _beginthread */
 #include <tchar.h>
 
@@ -289,6 +309,9 @@ typedef unsigned long u32;
 #include <arpa/inet.h>   /* inet_ntoa */
 #include <sys/time.h>    /* select */
 #include <sys/ioctl.h>   /* ioctl */
+#ifdef HAVE_SYS_UN_H
+#include <sys/un.h>
+#endif
 #include <netinet/tcp.h>
 #include <netdb.h>
 #ifndef INADDR_ANY
@@ -330,6 +353,7 @@ extern char *sys_errlist[];
 #endif /* IP_TRANSPARENT */
 #ifdef HAVE_LINUX_NETFILTER_IPV4_H
 #include <limits.h>
+#include <linux/types.h>
 #include <linux/netfilter_ipv4.h>
 #endif /* HAVE_LINUX_NETFILTER_IPV4_H */
 #endif /* __linux__ */
@@ -338,14 +362,16 @@ extern char *sys_errlist[];
 
 /**************************************** OpenSSL headers */
 
-#ifdef HAVE_OPENSSL
-
 #define OPENSSL_THREAD_DEFINES
 #include <openssl/opensslconf.h>
-#if !defined(OPENSSL_THREADS) && defined(USE_PTHREAD)
+#if defined(USE_PTHREAD) && !(defined(OPENSSL_THREADS) || \
+    (OPENSSL_VERSION_NUMBER<0x0090700fL && defined(THREADS)))
 #error OpenSSL library compiled without thread support
 #endif /* !OPENSSL_THREADS && USE_PTHREAD */
 
+/* OpenSSL 0.9.6 comp.h needs ZLIB macro to declare COMP_zlib() */
+#define ZLIB
+
 #include <openssl/lhash.h>
 #include <openssl/ssl.h>
 #include <openssl/err.h>
@@ -364,29 +390,23 @@ extern char *sys_errlist[];
 #endif
 #endif /* HAVE_OSSL_ENGINE_H */
 
+#ifdef HAVE_OSSL_OCSP_H
 #include <openssl/ocsp.h>
+#endif /* HAVE_OSSL_OCSP_H */
 
 #ifdef USE_FIPS
 #include <openssl/fips.h>
 #include <openssl/fips_rand.h>
 #endif /* USE_FIPS */
 
-#if OPENSSL_VERSION_NUMBER<0x10000000
+#if OPENSSL_VERSION_NUMBER<0x0090800fL
+#define OPENSSL_NO_ECDH
+#endif /* OpenSSL version < 0.8.0 */
+
+#if OPENSSL_VERSION_NUMBER<0x10000000L
 #define OPENSSL_NO_TLSEXT
 #endif /* OpenSSL version < 1.0.0 */
 
-#else /* HAVE_OPENSSL */
-
-#include <lhash.h>
-#include <ssl.h>
-#include <err.h>
-#include <crypto.h>
-#include <md4.h>
-#include <des.h>
-#define OPENSSL_NO_TLSEXT
-
-#endif /* HAVE_OPENSSL */
-
 /**************************************** other defines */
 
 /* change all non-printable characters to '.' */
@@ -398,14 +418,6 @@ extern char *sys_errlist[];
     do {unsigned char *p; for(p=(s); *p; p++) \
         if(!isalnum((int)*p)) *p='.';} while(0)
 
-/* some definitions for IPv6 support */
-#if defined(USE_IPv6)
-#define addr_len(x) ((x).sa.sa_family==AF_INET ? \
-    sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6))
-#else
-#define addr_len(x) (sizeof(struct sockaddr_in))
-#endif
-
 /* always use IPv4 defaults! */
 #define DEFAULT_LOOPBACK "127.0.0.1"
 #define DEFAULT_ANY "0.0.0.0"
