@@ -38,11 +38,10 @@
 #include "common.h"
 #include "prototypes.h"
 
-/* \n is not a character expected in the string */
-#define LINE "%[^\n]"
 #define isprefix(a, b) (strncasecmp((a), (b), strlen(b))==0)
 
 /* protocol-specific function prototypes */
+static void proxy_server(CLI *c);
 static void cifs_client(CLI *);
 static void cifs_server(CLI *);
 static void pgsql_client(CLI *);
@@ -55,61 +54,125 @@ static void imap_client(CLI *);
 static void imap_server(CLI *);
 static void nntp_client(CLI *);
 static void connect_client(CLI *);
+
+#if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
 static void ntlm(CLI *);
-#ifndef OPENSSL_NO_MD4
 static char *ntlm1();
 static char *ntlm3(char *, char *, char *);
 static void crypt_DES(DES_cblock, DES_cblock, DES_cblock);
 #endif
 static char *base64(int, char *, int);
 
-void negotiate(CLI *c) {
-    if(!c->opt->protocol)
-        return; /* no protocol negotiations */
+/**************************************** framework */
 
-    s_log(LOG_NOTICE, "Negotiations for %s (%s side) started", c->opt->protocol,
-        c->opt->option.client ? "client" : "server");
+typedef void (*FUNCTION)(CLI *);
 
-    if(c->opt->option.client) {
-        if(!strcmp(c->opt->protocol, "cifs"))
-            cifs_client(c);
-        else if(!strcmp(c->opt->protocol, "smtp"))
-            smtp_client(c);
-        else if(!strcmp(c->opt->protocol, "pop3"))
-            pop3_client(c);
-        else if(!strcmp(c->opt->protocol, "imap"))
-            imap_client(c);
-        else if(!strcmp(c->opt->protocol, "nntp"))
-            nntp_client(c);
-        else if(!strcmp(c->opt->protocol, "connect"))
-            connect_client(c);
-        else if(!strcmp(c->opt->protocol, "pgsql"))
-            pgsql_client(c);
-        else {
-            s_log(LOG_ERR, "Protocol %s not supported in client mode",
-                c->opt->protocol);
-            longjmp(c->err, 1);
-        }
-    } else {
-        if(!strcmp(c->opt->protocol, "cifs"))
-            cifs_server(c);
-        else if(!strcmp(c->opt->protocol, "smtp"))
-            smtp_server(c);
-        else if(!strcmp(c->opt->protocol, "pop3"))
-            pop3_server(c);
-        else if(!strcmp(c->opt->protocol, "imap"))
-            imap_server(c);
-        else if(!strcmp(c->opt->protocol, "pgsql"))
-            pgsql_server(c);
-        else {
-            s_log(LOG_ERR, "Protocol %s not supported in server mode",
-                c->opt->protocol);
-            longjmp(c->err, 1);
-        }
-    }
-    s_log(LOG_NOTICE, "Protocol negotiations succeeded");
+static const struct {
+    char *name;
+    FUNCTION f[2][2];
+} protocols[]={
+    {"proxy",   {{proxy_server, NULL},           {NULL, NULL}}},
+    {"cifs",    {{cifs_server,  cifs_client},    {NULL, NULL}}},
+    {"pgsql",   {{pgsql_server, pgsql_client},   {NULL, NULL}}},
+    {"smtp",    {{smtp_server,  smtp_client},    {NULL, NULL}}},
+    {"pop3",    {{pop3_server,  pop3_client},    {NULL, NULL}}},
+    {"imap",    {{imap_server,  imap_client},    {NULL, NULL}}},
+    {"nntp",    {{NULL,         nntp_client},    {NULL, NULL}}},
+    {"connect", {{NULL,         connect_client}, {NULL, NULL}}},
+    {NULL,      {{NULL,         NULL},           {NULL, NULL}}}
+};
+
+int find_protocol_id(const char *name) {
+    int id;
+
+    for(id=0; protocols[id].name; ++id)
+        if(!strcmp(name, protocols[id].name))
+            return id;
+    return -1;
 }
 
+void protocol(CLI *c, const int phase) {
+    const int id=c->opt->protocol, mode=(unsigned int)c->opt->option.client;
+
+    if(id<0)
+        return;
+    if(!protocols[id].f[phase][mode]) {
+        s_log(protocols[id].f[phase][mode^1] ? LOG_NOTICE : LOG_DEBUG,
+            "No %s-SSL %s-mode %s protocol negotiation supported",
+            phase ? "post" : "pre", mode ? "client" : "server",
+            protocols[id].name);
+        return;
+    }
+    s_log(LOG_INFO, "%s-SSL %s-mode %s protocol negotiations started",
+        phase ? "Post" : "Pre", mode ? "client" : "server",
+        protocols[id].name);
+    protocols[id].f[phase][mode](c);
+    s_log(LOG_INFO, "%s-SSL %s-mode %s protocol negotiations succeeded",
+        phase ? "Post" : "Pre", mode ? "client" : "server",
+        protocols[id].name);
+}
+
+/**************************************** proxy */
+
+/*
+ * PROXY protocol: http://haproxy.1wt.eu/download/1.5/doc/proxy-protocol.txt
+ * this is a protocol client support for stunnel acting as an SSL server 
+ * I don't think anything else is useful, but feel free to discuss on the
+ * stunnel-users mailing list if you disagree
+ */
+
+/* IP address textual representation length */
+/* 1234:6789:1234:6789:1234:6789:1234:6789 -> 40 chars with '\0' */
+#define IP_LEN 40
+#define PORT_LEN 6
+
+static void proxy_server(CLI *c) {
+    SOCKADDR_UNION addr;
+    socklen_t addrlen;
+    char src_host[IP_LEN], dst_host[IP_LEN];
+    char src_port[PORT_LEN], dst_port[PORT_LEN], *proto;
+    int err;
+
+    addrlen=sizeof addr;
+    if(getpeername(c->local_rfd.fd, &addr.sa, &addrlen)) {
+        sockerror("getpeername");
+        longjmp(c->err, 1);
+    }
+    err=getnameinfo(&addr.sa, addr_len(&addr), src_host, IP_LEN,
+        src_port, PORT_LEN, NI_NUMERICHOST|NI_NUMERICSERV);
+    if(err) {
+        s_log(LOG_ERR, "getnameinfo: %s", s_gai_strerror(err));
+        longjmp(c->err, 1);
+    }
+
+    addrlen=sizeof addr;
+    if(getsockname(c->local_rfd.fd, &addr.sa, &addrlen)) {
+        sockerror("getsockname");
+        longjmp(c->err, 1);
+    }
+    err=getnameinfo(&addr.sa, addr_len(&addr), dst_host, IP_LEN,
+        dst_port, PORT_LEN, NI_NUMERICHOST|NI_NUMERICSERV);
+    if(err) {
+        s_log(LOG_ERR, "getnameinfo: %s", s_gai_strerror(err));
+        longjmp(c->err, 1);
+    }
+
+    switch(addr.sa.sa_family) {
+    case AF_INET:
+        proto="TCP4";
+        break;
+    case AF_INET6:
+        proto="TCP6";
+        break;
+    default: /* AF_UNIX */
+        proto="UNKNOWN";
+    }
+    fdprintf(c, c->remote_fd.fd, "PROXY %s %s %s %s %s",
+        proto, src_host, dst_host, src_port, dst_port);
+}
+
+/**************************************** cifs */
+
 static void cifs_client(CLI *c) {
     u8 buffer[5];
     u8 request_dummy[4] = {0x81, 0, 0, 0}; /* a zero-length request */
@@ -152,7 +215,9 @@ static void cifs_server(CLI *c) {
     write_blocking(c, c->local_wfd.fd, response_use_ssl, 5);
 }
 
-/* http://www.postgresql.org/docs/8.3/static/protocol-flow.html#AEN73013 */
+/**************************************** pgsql */
+
+/* http://www.postgresql.org/docs/8.3/static/protocol-flow.html#AEN73982 */
 u8 ssl_request[8]={0, 0, 0, 8, 0x04, 0xd2, 0x16, 0x2f};
 
 static void pgsql_client(CLI *c) {
@@ -180,17 +245,19 @@ static void pgsql_server(CLI *c) {
     write_blocking(c, c->local_wfd.fd, ssl_ok, sizeof ssl_ok);
 }
 
+/**************************************** smtp */
+
 static void smtp_client(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
     do { /* copy multiline greeting */
-        fdgetline(c, c->remote_fd.fd, line);
+        line=fdgetline(c, c->remote_fd.fd);
         fdputline(c, c->local_wfd.fd, line);
     } while(isprefix(line, "220-"));
 
     fdputline(c, c->remote_fd.fd, "EHLO localhost");
     do { /* skip multiline reply */
-        fdgetline(c, c->remote_fd.fd, line);
+        line=fdgetline(c, c->remote_fd.fd);
     } while(isprefix(line, "250-"));
     if(!isprefix(line, "250 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 1425 compliant");
@@ -199,7 +266,7 @@ static void smtp_client(CLI *c) {
 
     fdputline(c, c->remote_fd.fd, "STARTTLS");
     do { /* skip multiline reply */
-        fdgetline(c, c->remote_fd.fd, line);
+        line=fdgetline(c, c->remote_fd.fd);
     } while(isprefix(line, "220-"));
     if(!isprefix(line, "220 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 2487 compliant");
@@ -208,11 +275,11 @@ static void smtp_client(CLI *c) {
 }
 
 static void smtp_server(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
-    s_poll_init(&c->fds);
-    s_poll_add(&c->fds, c->local_rfd.fd, 1, 0);
-    switch(s_poll_wait(&c->fds, 0, 200)) { /* wait up to 200ms */
+    s_poll_init(c->fds);
+    s_poll_add(c->fds, c->local_rfd.fd, 1, 0);
+    switch(s_poll_wait(c->fds, 0, 200)) { /* wait up to 200ms */
     case 0: /* fd not ready to read */
         s_log(LOG_DEBUG, "RFC 2487 detected");
         break;
@@ -224,20 +291,20 @@ static void smtp_server(CLI *c) {
         longjmp(c->err, 1);
     }
 
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "220")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdprintf(c, c->local_wfd.fd, "%s + stunnel", line);
-    fdgetline(c, c->local_rfd.fd, line);
+    line=fdgetline(c, c->local_rfd.fd);
     if(!isprefix(line, "EHLO ")) {
         s_log(LOG_ERR, "Unknown client EHLO");
         longjmp(c->err, 1);
     }
     fdprintf(c, c->local_wfd.fd, "250-%s Welcome", line);
     fdputline(c, c->local_wfd.fd, "250 STARTTLS");
-    fdgetline(c, c->local_rfd.fd, line);
+    line=fdgetline(c, c->local_rfd.fd);
     if(!isprefix(line, "STARTTLS")) {
         s_log(LOG_ERR, "STARTTLS expected");
         longjmp(c->err, 1);
@@ -245,17 +312,19 @@ static void smtp_server(CLI *c) {
     fdputline(c, c->local_wfd.fd, "220 Go ahead");
 }
 
-static void pop3_client(CLI *c) {
-    char line[STRLEN];
+/**************************************** pop3 */
 
-    fdgetline(c, c->remote_fd.fd, line);
+static void pop3_client(CLI *c) {
+    char *line;
+
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "+OK ")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdputline(c, c->local_wfd.fd, line);
     fdputline(c, c->remote_fd.fd, "STLS");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "+OK ")) {
         s_log(LOG_ERR, "Server does not support TLS");
         longjmp(c->err, 1);
@@ -263,14 +332,14 @@ static void pop3_client(CLI *c) {
 }
 
 static void pop3_server(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     fdprintf(c, c->local_wfd.fd, "%s + stunnel", line);
-    fdgetline(c, c->local_rfd.fd, line);
+    line=fdgetline(c, c->local_rfd.fd);
     if(isprefix(line, "CAPA")) { /* client wants RFC 2449 extensions */
         fdputline(c, c->local_wfd.fd, "-ERR Stunnel does not support capabilities");
-        fdgetline(c, c->local_rfd.fd, line);
+        line=fdgetline(c, c->local_rfd.fd);
     }
     if(!isprefix(line, "STLS")) {
         s_log(LOG_ERR, "Client does not want TLS");
@@ -279,17 +348,19 @@ static void pop3_server(CLI *c) {
     fdputline(c, c->local_wfd.fd, "+OK Stunnel starts TLS negotiation");
 }
 
-static void imap_client(CLI *c) {
-    char line[STRLEN];
+/**************************************** imap */
 
-    fdgetline(c, c->remote_fd.fd, line);
+static void imap_client(CLI *c) {
+    char *line;
+
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdputline(c, c->local_wfd.fd, line);
     fdputline(c, c->remote_fd.fd, "stunnel STARTTLS");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "stunnel OK")) {
         fdputline(c, c->local_wfd.fd,
             "* BYE stunnel: Server does not support TLS");
@@ -299,11 +370,11 @@ static void imap_client(CLI *c) {
 }
 
 static void imap_server(CLI *c) {
-    char line[STRLEN], id[STRLEN], *tail, *capa;
+    char *line, *id, *tail, *capa;
  
-    s_poll_init(&c->fds);
-    s_poll_add(&c->fds, c->local_rfd.fd, 1, 0);
-    switch(s_poll_wait(&c->fds, 0, 200)) {
+    s_poll_init(c->fds);
+    s_poll_add(c->fds, c->local_rfd.fd, 1, 0);
+    switch(s_poll_wait(c->fds, 0, 200)) {
     case 0: /* fd not ready to read */
         s_log(LOG_DEBUG, "RFC 2595 detected");
         break;
@@ -316,7 +387,7 @@ static void imap_server(CLI *c) {
     }
 
     /* process server welcome and send it to client */
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
@@ -329,9 +400,9 @@ static void imap_server(CLI *c) {
     fdprintf(c, c->local_wfd.fd, "%s (stunnel)", line);
 
     while(1) { /* process client commands */
-        fdgetline(c, c->local_rfd.fd, line);
+        line=fdgetline(c, c->local_rfd.fd);
         /* split line into id and tail */
-        safecopy(id, line);
+        id=str_dup(line);
         tail=strchr(id, ' ');
         if(!tail)
             break;
@@ -343,7 +414,7 @@ static void imap_server(CLI *c) {
             return; /* success */
         } else if(isprefix(tail, "CAPABILITY")) {
             fdputline(c, c->remote_fd.fd, line); /* send it to server */
-            fdgetline(c, c->remote_fd.fd, line); /* get the capabilites */
+            line=fdgetline(c, c->remote_fd.fd); /* get the capabilites */
             if(*line=='*') {
                 /* 
                  * append STARTTLS
@@ -352,7 +423,7 @@ static void imap_server(CLI *c) {
                  * LOGIN would fail as "unexpected command", anyway
                  */
                 fdprintf(c, c->local_wfd.fd, "%s STARTTLS", line);
-                fdgetline(c, c->remote_fd.fd, line); /* next line */
+                line=fdgetline(c, c->remote_fd.fd); /* next line */
             }
             fdputline(c, c->local_wfd.fd, line); /* forward to the client */
             tail=strchr(line, ' ');
@@ -375,31 +446,35 @@ static void imap_server(CLI *c) {
     }    
     /* clean server shutdown */
     fdputline(c, c->remote_fd.fd, "stunnel LOGOUT");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(*line=='*')
-        fdgetline(c, c->remote_fd.fd, line);
+        line=fdgetline(c, c->remote_fd.fd);
     longjmp(c->err, 2); /* don't reset */
 }
 
-static void nntp_client(CLI *c) {
-    char line[STRLEN];
+/**************************************** nntp */
 
-    fdgetline(c, c->remote_fd.fd, line);
+static void nntp_client(CLI *c) {
+    char *line;
+
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "200 ") && !isprefix(line, "201 ")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdputline(c, c->local_wfd.fd, line);
     fdputline(c, c->remote_fd.fd, "STARTTLS");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "382 ")) {
         s_log(LOG_ERR, "Server does not support TLS");
         longjmp(c->err, 1);
     }
 }
 
+/**************************************** connect */
+
 static void connect_client(CLI *c) {
-    char line[STRLEN], *encoded;
+    char *line, *encoded;
 
     if(!c->opt->protocol_host) {
         s_log(LOG_ERR, "protocolHost not specified");
@@ -410,33 +485,44 @@ static void connect_client(CLI *c) {
     fdprintf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
     if(c->opt->protocol_username && c->opt->protocol_password) {
         if(!strcasecmp(c->opt->protocol_authentication, "NTLM")) {
+#if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
             ntlm(c);
+#else
+            s_log(LOG_ERR, "NTLM authentication is not available");
+            longjmp(c->err, 1);
+#endif
         } else { /* basic authentication */
-            safecopy(line, c->opt->protocol_username);
-            safeconcat(line, ":");
-            safeconcat(line, c->opt->protocol_password);
+            line=str_printf("%s:%s",
+                c->opt->protocol_username, c->opt->protocol_password);
             encoded=base64(1, line, strlen(line));
-            safecopy(line, encoded);
-            free(encoded);
+            str_free(line);
+            if(!encoded) {
+                s_log(LOG_ERR, "Base64 encoder failed");
+                longjmp(c->err, 1);
+            }
             fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: basic %s",
-                line);
+                encoded);
+            str_free(encoded);
         }
     }
     fdputline(c, c->remote_fd.fd, ""); /* empty line */
-    fdgetline(c, c->remote_fd.fd, line);
-    if(line[9]!='2') { /* "HTTP/1.0 200 Connection established" */
+    line=fdgetline(c, c->remote_fd.fd);
+    if(strlen(line)<12 || line[9]!='2') {
+        /* not "HTTP/1.0 200 Connection established" */
         s_log(LOG_ERR, "CONNECT request rejected");
         do { /* read all headers */
-            fdgetline(c, c->remote_fd.fd, line);
+            line=fdgetline(c, c->remote_fd.fd);
         } while(*line);
         longjmp(c->err, 1);
     }
     s_log(LOG_INFO, "CONNECT request accepted");
     do {
-        fdgetline(c, c->remote_fd.fd, line); /* read all headers */
+        line=fdgetline(c, c->remote_fd.fd); /* read all headers */
     } while(*line);
 }
 
+#if !defined(OPENSSL_NO_MD4) && OPENSSL_VERSION_NUMBER>=0x0090700fL
+
 /* 
  * NTLM code is based on the following documentation:
  * http://davenport.sourceforge.net/ntlm.html
@@ -446,34 +532,41 @@ static void connect_client(CLI *c) {
 #define s_min(a, b) ((a)>(b)?(b):(a))
 
 static void ntlm(CLI *c) {
-#ifndef OPENSSL_NO_MD4
-    char line[STRLEN], *encoded;
-    char buf[BUFSIZ], ntlm2[STRLEN];
-    long content_length;
+    char *line, buf[BUFSIZ], *ntlm1_txt, *ntlm2_txt, *ntlm3_txt;
+    long content_length=0; /* no HTTP content */
 
     /* send Proxy-Authorization (phase 1) */
     fdprintf(c, c->remote_fd.fd, "Proxy-Connection: keep-alive");
-    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm1());
+    ntlm1_txt=ntlm1();
+    if(!ntlm1_txt) {
+        s_log(LOG_ERR, "Proxy-Authenticate: Failed to build NTLM request");
+        longjmp(c->err, 1);
+    }
+    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm1_txt);
+    str_free(ntlm1_txt);
     fdputline(c, c->remote_fd.fd, ""); /* empty line */
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
 
     /* receive Proxy-Authenticate (phase 2) */
     if(line[9]!='4' || line[10]!='0' || line[11]!='7') { /* code 407 */
         s_log(LOG_ERR, "NTLM authorization request rejected");
         do { /* read all headers */
-            fdgetline(c, c->remote_fd.fd, line);
+            line=fdgetline(c, c->remote_fd.fd);
         } while(*line);
         longjmp(c->err, 1);
     }
-    *ntlm2='\0';
-    content_length=0; /* no HTTP content */
+    ntlm2_txt=NULL;
     do { /* read all headers */
-        fdgetline(c, c->remote_fd.fd, line);
+        line=fdgetline(c, c->remote_fd.fd);
         if(isprefix(line, "Proxy-Authenticate: NTLM "))
-            safecopy(ntlm2, line+25);
+            ntlm2_txt=str_dup(line+25);
         else if(isprefix(line, "Content-Length: "))
             content_length=atol(line+16);
     } while(*line);
+    if(!ntlm2_txt) { /* no Proxy-Authenticate: NTLM header */
+        s_log(LOG_ERR, "Proxy-Authenticate: NTLM header not found");
+        longjmp(c->err, 1);
+    }
 
     /* read and ignore HTTP content (if any) */
     while(content_length) {
@@ -484,18 +577,16 @@ static void ntlm(CLI *c) {
     /* send Proxy-Authorization (phase 3) */
     fdprintf(c, c->remote_fd.fd, "CONNECT %s HTTP/1.1", c->opt->protocol_host);
     fdprintf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
-    encoded=ntlm3(c->opt->protocol_username, c->opt->protocol_password, ntlm2);
-    safecopy(line, encoded);
-    free(encoded);
-    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", line);
-#else
-    s_log(LOG_ERR, "NTLM authentication is not available");
-    longjmp(c->err, 1);
-#endif
+    ntlm3_txt=ntlm3(c->opt->protocol_username, c->opt->protocol_password, ntlm2_txt);
+    str_free(ntlm2_txt);
+    if(!ntlm3_txt) {
+        s_log(LOG_ERR, "Proxy-Authenticate: Failed to build NTLM response");
+        longjmp(c->err, 1);
+    }
+    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm3_txt);
+    str_free(ntlm3_txt);
 }
 
-#ifndef OPENSSL_NO_MD4
-
 static char *ntlm1() {
     char phase1[16];
 
@@ -543,13 +634,15 @@ static char *ntlm3(char *username, char *password, char *phase2) {
 
     /* decode challenge and calculate response */
     decoded=base64(0, phase2, strlen(phase2)); /* decode */
+    if(!decoded)
+        return NULL;
     crypt_DES((unsigned char *)phase3+64,
         (unsigned char *)decoded+24, md4_hash);
     crypt_DES((unsigned char *)phase3+72,
         (unsigned char *)decoded+24, md4_hash+7);
     crypt_DES((unsigned char *)phase3+80,
         (unsigned char *)decoded+24, md4_hash+14);
-    free(decoded);
+    str_free(decoded);
 
     strncpy(phase3+88, username, sizeof phase3-88);
 
@@ -582,10 +675,17 @@ static void crypt_DES(DES_cblock dst, const_DES_cblock src, DES_cblock hash) {
 static char *base64(int encode, char *in, int len) {
     BIO *bio, *b64;
     char *out;
+    int n;
 
     b64=BIO_new(BIO_f_base64());
+    if(!b64)
+        return NULL;
     BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
     bio=BIO_new(BIO_s_mem());
+    if(!bio) {
+        str_free(b64);
+        return NULL;
+    }
     if(encode)
         bio=BIO_push(b64, bio);
     BIO_write(bio, in, len);
@@ -596,15 +696,16 @@ static char *base64(int encode, char *in, int len) {
     } else {
         bio=BIO_push(b64, bio);
     }
-    len=BIO_pending(bio);
+    n=BIO_pending(bio);
     /* 32 bytes as a safety precaution for passing decoded data to crypt_DES */
-    /* len+1 to get null-terminated string on encode */
-    out=calloc(len<32?32:len+1, 1);
-    if(!out) {
-        s_log(LOG_ERR, "Fatal memory allocation error");
-        die(2);
+    /* n+1 to get null-terminated string on encode */
+    out=str_alloc(n<32?32:n+1);
+    n=BIO_read(bio, out, n);
+    if(n<0) {
+        BIO_free_all(bio);
+        str_free(out);
+        return NULL;
     }
-    BIO_read(bio, out, len);
     BIO_free_all(bio);
     return out;
 }
