@@ -47,10 +47,12 @@
 /* no need for critical sections */
 
 void enter_critical_section(SECTION_CODE i) {
+    (void)i; /* skip warning about unused parameter */
     /* empty */
 }
 
 void leave_critical_section(SECTION_CODE i) {
+    (void)i; /* skip warning about unused parameter */
     /* empty */
 }
 
@@ -76,7 +78,6 @@ void leave_critical_section(SECTION_CODE i) {
 /* first context on the ready list is the active context */
 CONTEXT *ready_head=NULL, *ready_tail=NULL;         /* ready to execute */
 CONTEXT *waiting_head=NULL, *waiting_tail=NULL;     /* waiting on poll() */
-int next_id=1;
 
 unsigned long stunnel_process_id(void) {
     return (unsigned long)getpid();
@@ -86,70 +87,89 @@ unsigned long stunnel_thread_id(void) {
     return ready_head ? ready_head->id : 0;
 }
 
-static CONTEXT *new_context(int stack_size) {
+static CONTEXT *new_context(void) {
+    static int next_id=1;
     CONTEXT *context;
 
     /* allocate and fill the CONTEXT structure */
-    context=malloc(sizeof(CONTEXT));
+    context=str_alloc(sizeof(CONTEXT));
     if(!context) {
         s_log(LOG_ERR, "Unable to allocate CONTEXT structure");
         return NULL;
     }
-    context->stack=malloc(stack_size);
-    if(!context->stack) {
-        s_log(LOG_ERR, "Unable to allocate CONTEXT stack");
-        return NULL;
-    }
+    str_detach(context);
     context->id=next_id++;
     context->fds=NULL;
     context->ready=0;
-    /* some manuals claim that initialization of context structure is required */
-    if(getcontext(&context->context)<0) {
-        free(context->stack);
-        free(context);
-        ioerror("getcontext");
-        return NULL;
-    }
-    context->context.uc_link=NULL; /* it should never happen */
-#if defined(__sgi) || ARGC==2 /* obsolete ss_sp semantics */
-    context->context.uc_stack.ss_sp=context->stack+stack_size-8;
-#else
-    context->context.uc_stack.ss_sp=context->stack;
-#endif
-    context->context.uc_stack.ss_size=stack_size;
-    context->context.uc_stack.ss_flags=0;
 
-    /* attach to the tail of the ready queue */
+    /* append to the tail of the ready queue */
     context->next=NULL;
     if(ready_tail)
         ready_tail->next=context;
     ready_tail=context;
     if(!ready_head)
         ready_head=context;
+
     return context;
 }
 
 void sthreads_init(void) {
     /* create the first (listening) context and put it in the running queue */
-    if(!new_context(DEFAULT_STACK_SIZE)) {
+    if(!new_context()) {
         s_log(LOG_ERR, "Unable create the listening context");
         die(1);
     }
+    /* no need to initialize ucontext_t structure here
+       it will be initialied with swapcontext() call */
 }
 
 int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     CONTEXT *context;
 
     (void)ls; /* this parameter is only used with USE_FORK */
+    
     s_log(LOG_DEBUG, "Creating a new context");
-    context=new_context(arg->opt->stack_size);
+    context=new_context();
     if(!context) {
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
     }
+
+    /* initialize context_t structure */
+    if(getcontext(&context->context)<0) {
+        str_free(context);
+        if(arg)
+            str_free(arg);
+        if(s>=0)
+            closesocket(s);
+        ioerror("getcontext");
+        return -1;
+    }
+    context->context.uc_link=NULL; /* stunnel does not use uc_link */
+
+    /* create stack */
+    context->stack=str_alloc(arg->opt->stack_size);
+    if(!context->stack) {
+        str_free(context);
+        if(arg)
+            str_free(arg);
+        if(s>=0)
+            closesocket(s);
+        s_log(LOG_ERR, "Unable to allocate stack");
+        return -1;
+    }
+    str_detach(context->stack);
+#if defined(__sgi) || ARGC==2 /* obsolete ss_sp semantics */
+    context->context.uc_stack.ss_sp=context->stack+arg->opt->stack_size-8;
+#else
+    context->context.uc_stack.ss_sp=context->stack;
+#endif
+    context->context.uc_stack.ss_size=arg->opt->stack_size;
+    context->context.uc_stack.ss_flags=0;
+
     s_log(LOG_DEBUG, "Context %ld created", context->id);
     makecontext(&context->context, (void(*)(void))cli, ARGC, arg);
     return 0;
@@ -172,6 +192,7 @@ unsigned long stunnel_thread_id(void) {
 }
 
 static void null_handler(int sig) {
+    (void)sig; /* skip warning about unused parameter */
     signal(SIGCHLD, null_handler);
 }
 
@@ -179,7 +200,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     switch(fork()) {
     case -1:    /* error */
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
@@ -191,7 +212,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
         _exit(0);
     default:    /* parent */
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
     }
@@ -213,11 +234,7 @@ void leave_critical_section(SECTION_CODE i) {
     pthread_mutex_unlock(stunnel_cs+i);
 }
 
-static void locking_callback(int mode, int type,
-#ifdef HAVE_OPENSSL
-    const /* callback definition has been changed in openssl 0.9.3 */
-#endif
-    char *file, int line) {
+static void locking_callback(int mode, int type, const char *file, int line) {
     (void)file; /* skip warning about unused parameter */
     (void)line; /* skip warning about unused parameter */
     if(mode&CRYPTO_LOCK)
@@ -236,9 +253,10 @@ static struct CRYPTO_dynlock_value *dyn_create_function(const char *file,
 
     (void)file; /* skip warning about unused parameter */
     (void)line; /* skip warning about unused parameter */
-    value=malloc(sizeof(struct CRYPTO_dynlock_value));
+    value=str_alloc(sizeof(struct CRYPTO_dynlock_value));
     if(!value)
         return NULL;
+    str_detach(value);
     pthread_mutex_init(&value->mutex, NULL);
     return value;
 }
@@ -258,7 +276,7 @@ static void dyn_destroy_function(struct CRYPTO_dynlock_value *value,
     (void)file; /* skip warning about unused parameter */
     (void)line; /* skip warning about unused parameter */
     pthread_mutex_destroy(&value->mutex);
-    free(value);
+    str_free(value);
 }
 
 unsigned long stunnel_process_id(void) {
@@ -289,40 +307,42 @@ void sthreads_init(void) {
 }
 
 int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
-    pthread_attr_t pth_attr;
     pthread_t thread;
-#ifdef HAVE_PTHREAD_SIGMASK
-    sigset_t newmask, oldmask;
+    pthread_attr_t pth_attr;
+    int error;
+#if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
+    /* Disabled on OS X due to strange problems on Mac OS X 10.5
+       it seems to restore signal mask somewhere (I couldn't find where)
+       effectively blocking signals after first accepted connection */
+    sigset_t new_set, old_set;
+#endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
 
     (void)ls; /* this parameter is only used with USE_FORK */
-    /* initialize attributes for creating new threads */
+
+#if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
+    /* the idea is that only the main thread handles all the signals with
+     * posix threads;  signals are blocked for any other thread */
+    sigfillset(&new_set);
+    pthread_sigmask(SIG_SETMASK, &new_set, &old_set); /* block signals */
+#endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
     pthread_attr_init(&pth_attr);
     pthread_attr_setdetachstate(&pth_attr, PTHREAD_CREATE_DETACHED);
     pthread_attr_setstacksize(&pth_attr, arg->opt->stack_size);
+    error=pthread_create(&thread, &pth_attr, cli, arg);
+    pthread_attr_destroy(&pth_attr);
+#if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
+    pthread_sigmask(SIG_SETMASK, &old_set, NULL); /* unblock signals */
+#endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
 
-    /* the idea is that only the main thread handles all the signals with
-     * posix threads;  signals are blocked for any other thread */
-    sigemptyset(&newmask);
-    sigaddset(&newmask, SIGCHLD);
-    sigaddset(&newmask, SIGTERM);
-    sigaddset(&newmask, SIGQUIT);
-    sigaddset(&newmask, SIGINT);
-    sigaddset(&newmask, SIGHUP);
-    pthread_sigmask(SIG_BLOCK, &newmask, &oldmask); /* block signals */
-#endif /* HAVE_PTHREAD_SIGMASK */
-    if(pthread_create(&thread, &pth_attr, cli, arg)) {
-#ifdef HAVE_PTHREAD_SIGMASK
-        pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
-#endif /* HAVE_PTHREAD_SIGMASK */
+    if(error) {
+        errno=error;
+        ioerror("pthread_create");
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
     }
-#ifdef HAVE_PTHREAD_SIGMASK
-    pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
-#endif /* HAVE_PTHREAD_SIGMASK */
     return 0;
 }
 
@@ -341,11 +361,7 @@ void leave_critical_section(SECTION_CODE i) {
     LeaveCriticalSection(stunnel_cs+i);
 }
 
-static void locking_callback(int mode, int type,
-#ifdef HAVE_OPENSSL
-    const /* callback definition has been changed in openssl 0.9.3 */
-#endif
-    char *file, int line) {
+static void locking_callback(int mode, int type, const char *file, int line) {
     (void)file; /* skip warning about unused parameter */
     (void)line; /* skip warning about unused parameter */
     if(mode&CRYPTO_LOCK)
@@ -364,9 +380,10 @@ static struct CRYPTO_dynlock_value *dyn_create_function(const char *file,
 
     (void)file; /* skip warning about unused parameter */
     (void)line; /* skip warning about unused parameter */
-    value=malloc(sizeof(struct CRYPTO_dynlock_value));
+    value=str_alloc(sizeof(struct CRYPTO_dynlock_value));
     if(!value)
         return NULL;
+    str_detach(value);
     InitializeCriticalSection(&value->mutex);
     return value;
 }
@@ -386,7 +403,7 @@ static void dyn_destroy_function(struct CRYPTO_dynlock_value *value,
     (void)file; /* skip warning about unused parameter */
     (void)line; /* skip warning about unused parameter */
     DeleteCriticalSection(&value->mutex);
-    free(value);
+    str_free(value);
 }
 
 unsigned long stunnel_process_id(void) {
@@ -421,7 +438,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     if((long)_beginthread((void(*)(void *))cli, arg->opt->stack_size, arg)==-1) {
         ioerror("_beginthread");
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
@@ -463,7 +480,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     if((long)_beginthread((void(*)(void *))cli, NULL, arg->opt->stack_size, arg)==-1L) {
         ioerror("_beginthread");
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
