@@ -48,7 +48,9 @@ static int add_dir_lookup(X509_STORE *, char *);
 static int verify_callback(int, X509_STORE_CTX *);
 static int cert_check(CLI *c, X509_STORE_CTX *, int);
 static int crl_check(CLI *c, X509_STORE_CTX *);
+#ifdef HAVE_OSSL_OCSP_H
 static int ocsp_check(CLI *c, X509_STORE_CTX *);
+#endif
 
 /* utility functions */
 static void log_time(const int, const char *, ASN1_TIME *);
@@ -57,18 +59,18 @@ static void log_time(const int, const char *, ASN1_TIME *);
 
 int verify_init(SERVICE_OPTIONS *section) {
     if(section->verify_level<0)
-        return 1; /* no certificate verification */
+        return 0; /* OK - no certificate verification */
 
-    if(section->verify_level>1 && !section->ca_file && !section->ca_dir) {
+    if(section->verify_level>=2 && !section->ca_file && !section->ca_dir) {
         s_log(LOG_ERR,
             "Either CApath or CAfile has to be used for authentication");
-        return 0;
+        return 1; /* FAILED */
     }
 
     section->revocation_store=X509_STORE_new();
     if(!section->revocation_store) {
         sslerror("X509_STORE_new");
-        return 0;
+        return 1; /* FAILED */
     }
 
     if(section->ca_file) {
@@ -77,15 +79,15 @@ int verify_init(SERVICE_OPTIONS *section) {
             s_log(LOG_ERR, "Error loading verify certificates from %s",
                 section->ca_file);
             sslerror("SSL_CTX_load_verify_locations");
-            return 0;
+            return 1; /* FAILED */
         }
         /* list of trusted CAs for the client to choose the right cert */
         SSL_CTX_set_client_CA_list(section->ctx,
             SSL_load_client_CA_file(section->ca_file));
         s_log(LOG_DEBUG, "Loaded verify certificates from %s",
             section->ca_file);
-        if(!load_file_lookup(section->revocation_store, section->ca_file))
-            return 0;
+        if(load_file_lookup(section->revocation_store, section->ca_file))
+            return 1; /* FAILED */
     }
 
     if(section->ca_dir) {
@@ -94,27 +96,28 @@ int verify_init(SERVICE_OPTIONS *section) {
             s_log(LOG_ERR, "Error setting verify directory to %s",
                 section->ca_dir);
             sslerror("SSL_CTX_load_verify_locations");
-            return 0;
+            return 1; /* FAILED */
         }
         s_log(LOG_DEBUG, "Verify directory set to %s", section->ca_dir);
         add_dir_lookup(section->revocation_store, section->ca_dir);
     }
 
     if(section->crl_file)
-        if(!load_file_lookup(section->revocation_store, section->crl_file))
-            return 0;
+        if(load_file_lookup(section->revocation_store, section->crl_file))
+            return 1; /* FAILED */
 
     if(section->crl_dir) {
         section->revocation_store->cache=0; /* don't cache CRLs */
         add_dir_lookup(section->revocation_store, section->crl_dir);
     }
 
-    SSL_CTX_set_verify(section->ctx, section->verify_level==SSL_VERIFY_NONE ?
-        SSL_VERIFY_PEER : section->verify_level, verify_callback);
+    SSL_CTX_set_verify(section->ctx, SSL_VERIFY_PEER |
+        (section->verify_level>=2 ? SSL_VERIFY_FAIL_IF_NO_PEER_CERT : 0),
+        verify_callback);
 
-    if(section->ca_dir && section->verify_use_only_my)
-        s_log(LOG_NOTICE, "Peer certificate location %s", section->ca_dir);
-    return 1; /* OK */
+    if(section->ca_dir && section->verify_level>=3)
+        s_log(LOG_INFO, "Peer certificate location %s", section->ca_dir);
+    return 0; /* OK */
 }
 
 static int load_file_lookup(X509_STORE *store, char *name) {
@@ -123,15 +126,15 @@ static int load_file_lookup(X509_STORE *store, char *name) {
     lookup=X509_STORE_add_lookup(store, X509_LOOKUP_file());
     if(!lookup) {
         sslerror("X509_STORE_add_lookup");
-        return 0;
+        return 1; /* FAILED */
     }
     if(!X509_LOOKUP_load_file(lookup, name, X509_FILETYPE_PEM)) {
         s_log(LOG_ERR, "Failed to load %s revocation lookup file", name);
         sslerror("X509_LOOKUP_load_file");
-        return 0;
+        return 1; /* FAILED */
     }
     s_log(LOG_DEBUG, "Loaded %s revocation lookup file", name);
-    return 1; /* OK */
+    return 0; /* OK */
 }
 
 static int add_dir_lookup(X509_STORE *store, char *name) {
@@ -140,15 +143,15 @@ static int add_dir_lookup(X509_STORE *store, char *name) {
     lookup=X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir());
     if(!lookup) {
         sslerror("X509_STORE_add_lookup");
-        return 0;
+        return 1; /* FAILED */
     }
     if(!X509_LOOKUP_add_dir(lookup, name, X509_FILETYPE_PEM)) {
         s_log(LOG_ERR, "Failed to add %s revocation lookup directory", name);
         sslerror("X509_LOOKUP_add_dir");
-        return 0;
+        return 1; /* FAILED */
     }
     s_log(LOG_DEBUG, "Added %s revocation lookup directory", name);
-    return 1; /* OK */
+    return 0; /* OK */
 }
 
 /**************************************** verify callback */
@@ -157,7 +160,7 @@ static int verify_callback(int preverify_ok, X509_STORE_CTX *callback_ctx) {
         /* our verify callback function */
     SSL *ssl;
     CLI *c;
-    char subject_name[STRLEN];
+    char *subject_name;
 
     /* retrieve application specific data */
     ssl=X509_STORE_CTX_get_ex_data(callback_ctx,
@@ -165,53 +168,77 @@ static int verify_callback(int preverify_ok, X509_STORE_CTX *callback_ctx) {
     c=SSL_get_ex_data(ssl, cli_index);
 
     /* certificate name for logging */
-    X509_NAME_oneline(X509_get_subject_name(callback_ctx->current_cert),
-        subject_name, STRLEN);
-    safestring(subject_name);
+    subject_name=X509_NAME_oneline(
+        X509_get_subject_name(callback_ctx->current_cert), NULL, 0);
 
     s_log(LOG_DEBUG, "Starting certificate verification: depth=%d, %s",
         callback_ctx->error_depth, subject_name);
     if(!cert_check(c, callback_ctx, preverify_ok)) {
         s_log(LOG_WARNING, "Certificate check failed: depth=%d, %s",
             callback_ctx->error_depth, subject_name);
+        OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
     if(!crl_check(c, callback_ctx)) {
         s_log(LOG_WARNING, "CRL check failed: depth=%d, %s",
             callback_ctx->error_depth, subject_name);
+        OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
+#ifdef HAVE_OSSL_OCSP_H
     if(c->opt->option.ocsp && !ocsp_check(c, callback_ctx)) {
         s_log(LOG_WARNING, "OCSP check failed: depth=%d, %s",
             callback_ctx->error_depth, subject_name);
+        OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
+#endif /* HAVE_OSSL_OCSP_H */
     /* errnum=X509_STORE_CTX_get_error(ctx); */
     s_log(LOG_NOTICE, "Certificate accepted: depth=%d, %s",
         callback_ctx->error_depth, subject_name);
+    OPENSSL_free(subject_name);
     return 1; /* accept connection */
 }
 
 /**************************************** certificate checking */
 
 static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx, int preverify_ok) {
-    X509_OBJECT ret;
+    X509_OBJECT obj;
+#if OPENSSL_VERSION_NUMBER>=0x0090700fL
+    ASN1_BIT_STRING *local_key, *peer_key;
+#endif
 
-    if(c->opt->verify_level==SSL_VERIFY_NONE) {
+    if(c->opt->verify_level<1) {
         s_log(LOG_INFO, "CERT: Verification not enabled");
         return 1; /* accept connection */
     }
     if(!preverify_ok) {
         /* remote site specified a certificate, but it's not correct */
-        s_log(LOG_WARNING, "CERT: Verification error: %s",
-            X509_verify_cert_error_string(callback_ctx->error));
-        return 0; /* reject connection */
+        if(c->opt->verify_level>=4 && callback_ctx->error_depth>0) {
+            s_log(LOG_INFO, "CERT: Invalid CA certificate ignored");
+            return 1; /* accept connection */
+        } else {    
+            s_log(LOG_WARNING, "CERT: Verification error: %s",
+                X509_verify_cert_error_string(callback_ctx->error));
+            return 0; /* reject connection */
+        }
     }
-    if(c->opt->verify_use_only_my && callback_ctx->error_depth==0 &&
-            X509_STORE_get_by_subject(callback_ctx, X509_LU_X509,
-                X509_get_subject_name(callback_ctx->current_cert), &ret)!=1) {
-        s_log(LOG_WARNING, "CERT: Certificate not found in local repository");
-        return 0; /* reject connection */
+    if(c->opt->verify_level>=3 && callback_ctx->error_depth==0) {
+        if(X509_STORE_get_by_subject(callback_ctx, X509_LU_X509,
+                X509_get_subject_name(callback_ctx->current_cert), &obj)!=1) {
+            s_log(LOG_WARNING, "CERT: Certificate not found in local repository");
+            return 0; /* reject connection */
+        }
+#if OPENSSL_VERSION_NUMBER>=0x0090700fL
+        peer_key=X509_get0_pubkey_bitstr(callback_ctx->current_cert);
+        local_key=X509_get0_pubkey_bitstr(obj.data.x509);
+        if(!peer_key || !local_key || peer_key->length!=local_key->length ||
+                memcmp(peer_key->data, local_key->data, local_key->length)) {
+            s_log(LOG_WARNING, "CERT: Public keys do not match");
+            return 0; /* reject connection */
+        }
+#endif
+        s_log(LOG_INFO, "CERT: Locally installed certificate matched");
     }
     return 1; /* accept connection */
 }
@@ -314,12 +341,13 @@ static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     return 1; /* accept connection */
 }
 
+#ifdef HAVE_OSSL_OCSP_H
+
 /**************************************** OCSP checking */
 /* TODO: check OCSP server specified in the certificate */
 
 static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     int error, retval=0;
-    SOCKADDR_UNION addr;
     X509 *cert;
     X509 *issuer=NULL;
     OCSP_CERTID *certID;
@@ -332,12 +360,11 @@ static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     int status, reason;
 
     /* connect specified OCSP server (responder) */
-    c->fd=s_socket(c->opt->ocsp_addr.addr[0].sa.sa_family, SOCK_STREAM, 0,
+    c->fd=s_socket(c->opt->ocsp_addr.sa.sa_family, SOCK_STREAM, 0,
         0, "OCSP: socket (auth_user)");
     if(c->fd<0)
         return 0; /* reject connection */
-    memcpy(&addr, &c->opt->ocsp_addr.addr[0], sizeof addr);
-    if(connect_blocking(c, &addr, addr_len(addr)))
+    if(connect_blocking(c, &c->opt->ocsp_addr, addr_len(&c->opt->ocsp_addr)))
         goto cleanup;
     s_log(LOG_DEBUG, "OCSP: server connected");
 
@@ -437,6 +464,8 @@ cleanup:
     return retval;
 }
 
+#endif /* HAVE_OSSL_OCSP_H */
+
 static void log_time(const int level, const char *txt, ASN1_TIME *t) {
     char *cp;
     BIO *bio;
@@ -445,14 +474,25 @@ static void log_time(const int level, const char *txt, ASN1_TIME *t) {
     if(!t)
         return;
     bio=BIO_new(BIO_s_mem());
+    if(!bio)
+        return;
     ASN1_TIME_print(bio, t);
     n=BIO_pending(bio);
-    cp=malloc(n+1);
+    cp=str_alloc(n+1);
+    if(!cp) {
+        BIO_free(bio);
+        return;
+    }
     n=BIO_read(bio, cp, n);
+    if(n<0) {
+        BIO_free(bio);
+        str_free(cp);
+        return;
+    }
     cp[n]='\0';
     BIO_free(bio);
     s_log(level, "%s: %s", txt, cp);
-    free(cp);
+    str_free(cp);
 }
 
 /* end of verify.c */
